	Subroutine PrsTarget
*
*	Routine to pick up strings which are search targets.
*	We allocate dynamic strings for each one as it is parsed,
*	and store the descriptors in a common block array.
*
*	Following suggestion of Pat Rankins, we check to see whether
*	all strings are non-alphabetic, and if so, force exact matching
*	because of its greater speed.
*
*	Following a further suggestion of his, a more reliable algorithm
*	is used to check for alphabetic strings, which copes with multi-
*	-national sets.
*
*	The local qualifiers which say which message components each
*	string applies to are also picked up here.

	Implicit None
	Include 'TRGCOM.INC'
	Include 'QUALCOM.INC'
	Character*128 Temp
	Integer Tlen,I
	Logical Ignore_Case,Swapd
	Record /DSCDEF1/ Ttarg
	Integer Status,Cli$Get_Value,Str$Copy_Dx,Str$Find_First_In_Set
	Integer CLI$Present,Cstatus
	character *128	alphabet
	integer		alen
	External CLI$_Absent,CLI$_Locneg,CLI$_locpres
	External MLSEARCH_TOOMNYTRG
	Integer Cflag

	Ntargets = 0
	CStatus = 1
	Ignore_Case = .True.
	If(.not.Exact)	Call Case_Sensitive_Chars( alphabet, alen, 1)

	Do While(CStatus)
	  CStatus = Cli$Get_Value('STRING',Temp,Tlen)
	  If(.Not. CStatus) Then
		Continue
	  Else If(Ntargets .ge. Maxtarget) Then
	    Call Lib$Signal(MLSEARCH_TOOMNYTRG,%val(1),%Val(Maxtarget))
	  Else
*
*	We have a string. Check for alphabeticness
*	If we don't want exact matching, uppercase the target.
*
	    If(.Not.Exact) Then
		Call Str$Upcase(Temp,Temp(1:Tlen))
	        If ( ignore_case )
     &		  ignore_case = (Str$Find_First_in_Set( Temp(1:Tlen),
     &						alphabet(:alen)) .eq. 0) 
	    Endif

	    Ntargets = Ntargets + 1
	    Call DynIni(Target(Ntargets))
	    Status = Str$Copy_Dx(Target(Ntargets),Temp(1:Tlen))
	    Call Stchek(Status)
*
*	Check for local qualifiers
*
	    Cflag = Sparts			!initial copy from global

	    Status = Cli$Present('INALL')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = -1
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = 0
	    Endif

	    Status = Cli$Present('INFROM')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = IOR(Cflag,S_From)
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = Iand(Cflag,NOT(S_From))
	    Endif

	    Status = Cli$Present('INTO')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = IOR(Cflag,S_To)
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = Iand(Cflag,NOT(S_To))
	    Endif

	    Status = Cli$Present('INCC')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = IOR(Cflag,S_CC)
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = Iand(Cflag,NOT(S_CC))
	    Endif

	    Status = Cli$Present('INSUBJECT')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = IOR(Cflag,S_Subject)
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = Iand(Cflag,NOT(S_Subject))
	    Endif

	    Status = Cli$Present('INTEXT')
	    If(Status .eq. %Loc(cli$_locpres)) Then
		Cflag = IOR(Cflag,S_Text)
	    Else If(Status .eq. %Loc(cli$_locneg)) Then
		Cflag = Iand(Cflag,NOT(S_Text))
	    Endif

	    Trg_flags(Ntargets) = Cflag

	  Endif

	Enddo

	If(Status .ne. %Loc(Cli$_Absent) .or. Ntargets .eq. 0)
     $		Call Stchek(Status)

	If(Ignore_Case) Exact = .True.
*
*	Now sort the targets in length order, longest first. This is to allow
*	highlighting to work correctly later on output.
*
	If(Ntargets .eq. 1) Return

	Swapd = .True.

	Do While (Swapd)
	 Swapd = .False.
	 Do I = 1,Ntargets-1
	  If(Target(I).Dsc$W_Maxstrlen .lt.
     $             Target(I+1).Dsc$W_Maxstrlen) Then
		Ttarg = Target(I)
		Target(I) = Target(I+1)
		Target(I+1) = Ttarg
		Swapd = .True.
	  Endif
	 Enddo
	Enddo

	Return

	End

      subroutine  Case_Sensitive_Chars ( string, length, typ_flag )
	!						Pat Rankin, May'89
	!   Create a string that contains all characters that are
	!   case sensitive [based on STR$UPCASE conversion behavior].
	!   (This is not quite the same as "alphabetic" characters.)
	!   Don't bother checking characters 0 through 63 because we
	!   assume that they are definitely non-alphabetic.
	!
	!   'typ_flag' indicates whether to collect upper-case (1),
	!   lower-case (2), or both (3) types of characters.
	!
	!   In the standard DEC-Multinational character set, there are
	!   55 case-sensitive 'letters', giving a string of length 110
	!   if both upper- and lower-case characters are requested.
	!
	implicit none
C     input:
	integer		typ_flag
C     output:
	character *(*)	string
	integer *2	length
C     local:
	character *(256-64) mixed,  upper	!(192)
	integer		i,  o,	ln
C     functions:
	intrinsic	CHAR,  LEN
*	subroutine	STR$UPCASE

C     make string containing all characters of interest
	do i = 64, 255		!skip 0..63 (they are non-alphabetic)
	    mixed(i-63:i-63) = CHAR(i)
	end do
C     convert it into uppercase
	call STR$UPCASE( upper, mixed)
C     compare upper-case string with mixed-case string, collecting difs
	ln = LEN(string)	!loop limit; reduce if both types wanted
	if ( (typ_flag.and.3).eq. 3 )  ln = ln - 1  !simplify bounds check
	o = 0			!output string length (so far)
	do i = 1, LEN(upper)	!loop through string
	    if ( upper(i:i) .ne. mixed(i:i) .and. o .lt. ln ) then
C	      got character (actually, pair of them) that's case sensitive
		if ( typ_flag.and.1 ) then		!want upper-case
		    o = o + 1			!increment output length
		    string(o:o) = upper(i:i)	!store character
		end if
		if ( (typ_flag.and.2).ne. 0 ) then	!want lower-case
		    o = o + 1			!increment output length
		    string(o:o) = mixed(i:i)	!store character
		end if
	    end if
	end do
C     done
	if ( o .lt. LEN(string) )  string(o+1:) = ' '	!blank fill
	length = o		!return length of output string
	return
      end !of Case_Sensitive_Chars

	Subroutine Dynini(Des)
*
*	Initialize a dynamic string descriptor
*
	Implicit None
	Include '($dscdef)'
	Record /DSCDEF1/ Des

	Des.DSC$W_MAXSTRLEN = 0
	Des.DSC$B_DTYPE = DSC$K_DTYPE_T
	Des.DSC$B_CLASS = DSC$K_CLASS_D
	Des.DSC$A_POINTER = 0

	Return
	End
